import inspect
from functools import partial
from typing import Optional, Callable, Tuple
from jsons._common_impl import get_class_name, META_ATTR, get_cls_from_str
from jsons._main_impl import load
from jsons.exceptions import SignatureMismatchError, UnfulfilledArgumentError


def default_object_deserializer(
        obj: dict,
        cls: type,
        key_transformer: Optional[Callable[[str], str]] = None,
        strict: bool = False,
        **kwargs) -> object:
    """
    Deserialize ``obj`` into an instance of type ``cls``. If ``obj`` contains
    keys with a certain case style (e.g. camelCase) that do not match the style
    of ``cls`` (e.g. snake_case), a key_transformer should be used (e.g.
    KEY_TRANSFORMER_SNAKECASE).
    :param obj: a serialized instance of ``cls``.
    :param cls: the type to which ``obj`` should be deserialized.
    :param key_transformer: a function that transforms the keys in order to
    match the attribute names of ``cls``.
    :param strict: deserialize in strict mode.
    :param kwargs: any keyword arguments that may be passed to the
    deserializers.
    :return: an instance of type ``cls``.
    """
    obj, kwargs = _check_and_transform_keys(obj, key_transformer, **kwargs)
    kwargs['strict'] = strict
    constructor_args = _get_constructor_args(obj, cls, **kwargs)
    remaining_attrs = _get_remaining_args(obj, cls, constructor_args,
                                          strict, kwargs['fork_inst'])
    instance = cls(**constructor_args)
    _set_remaining_attrs(instance, remaining_attrs, **kwargs)
    return instance


def _get_constructor_args(
        obj,
        cls,
        meta_hints,
        attr_getters=None,
        **kwargs) -> dict:
    # Loop through the signature of cls: the type we try to deserialize to. For
    # every required parameter, we try to get the corresponding value from
    # json_obj.
    signature_parameters = inspect.signature(cls.__init__).parameters
    attr_getters = dict(**(attr_getters or {}))
    value_for_attr_part = partial(_get_value_for_attr,
                                  obj=obj,
                                  cls=cls,
                                  meta_hints=meta_hints,
                                  attr_getters=attr_getters,
                                  **kwargs)
    args_gen = (value_for_attr_part(sig_key=sig_key, sig=sig) for sig_key, sig
                in signature_parameters.items() if sig_key != 'self')
    constructor_args_in_obj = {key: value for key, value in args_gen if key}
    return constructor_args_in_obj


def _get_value_for_attr(
        obj,
        cls,
        sig_key,
        sig,
        meta_hints,
        attr_getters,
        **kwargs):
    # Find a value for the attribute (with signature sig_key).
    result = None, None
    if obj and sig_key in obj:
        # This argument is in obj.
        result = sig_key, _get_value_from_obj(obj, sig, sig_key,
                                              meta_hints, **kwargs)
    elif sig_key in attr_getters:
        # There exists an attr_getter for this argument.
        attr_getter = attr_getters.pop(sig_key)
        result = sig_key, attr_getter()
    elif sig.default != inspect.Parameter.empty:
        # There is a default value for this argument.
        result = sig_key, sig.default
    elif sig.kind not in (inspect.Parameter.VAR_POSITIONAL,
                          inspect.Parameter.VAR_KEYWORD):
        # This argument is no *args or **kwargs and has no value.
        raise UnfulfilledArgumentError(
            'No value found for "{}"'.format(sig_key), sig_key, obj, cls)
    return result


def _get_value_from_obj(obj, sig, sig_key, meta_hints, **kwargs):
    # Obtain the value for the attribute with the given signature from the
    # given obj. Try to obtain the class of this attribute from the meta info
    # or from type hints.
    cls_key = '/{}'.format(sig_key)
    cls_from_meta = meta_hints.get(cls_key, None)
    new_hints = meta_hints
    arg_cls = None
    if cls_from_meta:
        arg_cls = get_cls_from_str(cls_from_meta, obj, kwargs['fork_inst'])
        # Rebuild the class hints: cls_key becomes the new root.
        new_hints = {
            key.replace(cls_key, '/'): meta_hints[key]
            for key in meta_hints
            if key != '/'
        }
    elif sig.annotation != inspect.Parameter.empty:
        arg_cls = sig.annotation
    value = load(obj[sig_key], arg_cls, meta_hints=new_hints, **kwargs)
    return value


def _set_remaining_attrs(instance,
                         remaining_attrs,
                         attr_getters=None,
                         **kwargs):
    # Set any remaining attributes on the newly created instance.
    attr_getters = attr_getters or {}
    for attr_name in remaining_attrs:
        loaded_attr = load(remaining_attrs[attr_name],
                           type(remaining_attrs[attr_name]),
                           **kwargs)
        try:
            setattr(instance, attr_name, loaded_attr)
        except AttributeError:
            pass  # This is raised when a @property does not have a setter.
    for attr_name, getter in attr_getters.items():
        setattr(instance, attr_name, getter())


def _check_and_transform_keys(obj: dict,
                              key_transformer: Optional[Callable[[str], str]],
                              **kwargs) -> Tuple[dict, dict]:
    if key_transformer:
        obj = {key_transformer(key): obj[key] for key in obj}
        kwargs = {
            **kwargs,
            'key_transformer': key_transformer
        }
    return obj, kwargs


def _get_remaining_args(obj: dict,
                        cls: type,
                        constructor_args: dict,
                        strict: bool,
                        fork_inst: type) -> dict:
    # Get the remaining args or raise if strict and the signature is unmatched.
    remaining_attrs = {attr_name: obj[attr_name] for attr_name in obj
                       if attr_name not in constructor_args
                       and attr_name != META_ATTR}
    if strict and remaining_attrs:
        unexpected_arg = list(remaining_attrs.keys())[0]
        err_msg = ('Type "{}" does not expect "{}"'
                   .format(get_class_name(cls, fork_inst=fork_inst),
                           unexpected_arg))
        raise SignatureMismatchError(err_msg, unexpected_arg, obj, cls)
    return remaining_attrs
